from pathlib import Path
from typing import Dict, Union

from . import Context, Props


def play(props: Dict[Union[str, Props], str] = None, **other_props: str) -> Context:
    """Play a sound with the specified props

    :param props:
        A mapping of :class:`Props` to values.

    :param other_props:
        :class:`Props` may also be passed as kwargs, where ``{Props.EVENT_ID: 'bell'}``
        is passed as ``event_id='bell'``.

    :return:
        The canberra.Context used to play the sound. `cancel()` may be called
        on this context to stop the playing sound.

    """
    ctx = Context()
    ctx.play(props, **other_props)
    return ctx


def play_file(filename: Union[str, Path], **other_props: str) -> Context:
    """Play the specified sound file

    :param filename:
        Path to the sound file to be played

    :param other_props:
        :class:`Props` passed as kwargs, where ``{Props.EVENT_ID: 'bell'}`` is
        passed as ``event_id='bell'``.

    :return:
        The :class:`canberra.Context` used to play the sound. `cancel()` may be
        called on this context to stop the playing sound.

    """
    filename = str(filename)
    return play(**other_props, media_filename=filename)



# distutils: language = c
# cython: language_level=3

from queue import Queue
from threading import Event, Thread
from typing import Any, Callable, Dict, Union

from cpython.object cimport PyObject
from cpython.pystate cimport PyGILState_STATE, PyGILState_Ensure, PyGILState_Release
from cpython.ref cimport Py_INCREF, Py_DECREF
from libc.stdint cimport uint32_t
from libc.stdlib cimport malloc, free
from posix.dlfcn cimport dlopen, RTLD_NOW, RTLD_GLOBAL, dlsym

from .constants import Props, Errors, NOTSET


cdef extern from 'canberra.h':
    ctypedef struct ca_context:
        pass

    ctypedef struct ca_proplist:
        pass


cdef:
    enum:
        CA_SUCCESS = 0
        CA_ERROR_NOTSUPPORTED = -1
        CA_ERROR_INVALID = -2
        CA_ERROR_STATE = -3
        CA_ERROR_OOM = -4
        CA_ERROR_NODRIVER = -5
        CA_ERROR_SYSTEM = -6
        CA_ERROR_CORRUPT = -7
        CA_ERROR_TOOBIG = -8
        CA_ERROR_NOTFOUND = -9
        CA_ERROR_DESTROYED = -10
        CA_ERROR_CANCELED = -11
        CA_ERROR_NOTAVAILABLE = -12
        CA_ERROR_ACCESS = -13
        CA_ERROR_IO = -14
        CA_ERROR_INTERNAL = -15
        CA_ERROR_DISABLED = -16
        CA_ERROR_FORKED = -17
        CA_ERROR_DISCONNECTED = -18

    ctypedef void (*ca_finish_callback_t)(ca_context *c, uint32_t id, int error_code, void *userdata);

    ctypedef int (*ca_proplist_create_t)(ca_proplist **p);
    ctypedef int (*ca_proplist_destroy_t)(ca_proplist *p);
    ctypedef int (*ca_proplist_sets_t)(ca_proplist *p, const char *key, const char *value);
    ctypedef int (*ca_proplist_setf_t)(ca_proplist *p, const char *key, const char *format, ...);
    ctypedef int (*ca_proplist_set_t)(ca_proplist *p, const char *key, const void *data, size_t nbytes);

    ctypedef int (*ca_context_create_t)(ca_context **c);
    ctypedef int (*ca_context_set_driver_t)(ca_context *c, const char *driver);
    ctypedef int (*ca_context_change_device_t)(ca_context *c, const char *device);
    ctypedef int (*ca_context_open_t)(ca_context *c);
    ctypedef int (*ca_context_destroy_t)(ca_context *c);
    ctypedef int (*ca_context_change_props_t)(ca_context *c, ...);
    ctypedef int (*ca_context_change_props_full_t)(ca_context *c, ca_proplist *p);
    ctypedef int (*ca_context_play_full_t)(ca_context *c, uint32_t id, ca_proplist *p, ca_finish_callback_t cb, void *userdata);
    ctypedef int (*ca_context_play_t)(ca_context *c, uint32_t id, ...);
    ctypedef int (*ca_context_cache_full_t)(ca_context *c, ca_proplist *p);
    ctypedef int (*ca_context_cache_t)(ca_context *c, ...);
    ctypedef int (*ca_context_cancel_t)(ca_context *c, uint32_t id);
    ctypedef int (*ca_context_playing_t)(ca_context *c, uint32_t id, int *playing);

    ctypedef const char *(*ca_strerror_t)(int code);


cdef:
    void *libcanberra = dlopen('libcanberra.so', RTLD_NOW | RTLD_GLOBAL)

    ca_proplist_create_t ca_proplist_create                     = <ca_proplist_create_t>dlsym(libcanberra, 'ca_proplist_create')
    ca_proplist_destroy_t ca_proplist_destroy                   = <ca_proplist_destroy_t>dlsym(libcanberra, 'ca_proplist_destroy')
    ca_proplist_sets_t ca_proplist_sets                         = <ca_proplist_sets_t>dlsym(libcanberra, 'ca_proplist_sets')
    ca_proplist_setf_t ca_proplist_setf                         = <ca_proplist_setf_t>dlsym(libcanberra, 'ca_proplist_setf')
    ca_proplist_set_t ca_proplist_set                           = <ca_proplist_set_t>dlsym(libcanberra, 'ca_proplist_set')
    ca_context_create_t ca_context_create                       = <ca_context_create_t>dlsym(libcanberra, 'ca_context_create')
    ca_context_set_driver_t ca_context_set_driver               = <ca_context_set_driver_t>dlsym(libcanberra, 'ca_context_set_driver')
    ca_context_change_device_t ca_context_change_device         = <ca_context_change_device_t>dlsym(libcanberra, 'ca_context_change_device')
    ca_context_open_t ca_context_open                           = <ca_context_open_t>dlsym(libcanberra, 'ca_context_open')
    ca_context_destroy_t ca_context_destroy                     = <ca_context_destroy_t>dlsym(libcanberra, 'ca_context_destroy')
    ca_context_change_props_t ca_context_change_props           = <ca_context_change_props_t>dlsym(libcanberra, 'ca_context_change_props')
    ca_context_change_props_full_t ca_context_change_props_full = <ca_context_change_props_full_t>dlsym(libcanberra, 'ca_context_change_props_full')
    ca_context_play_full_t ca_context_play_full                 = <ca_context_play_full_t>dlsym(libcanberra, 'ca_context_play_full')
    ca_context_play_t ca_context_play                           = <ca_context_play_t>dlsym(libcanberra, 'ca_context_play')
    ca_context_cache_full_t ca_context_cache_full               = <ca_context_cache_full_t>dlsym(libcanberra, 'ca_context_cache_full')
    ca_context_cache_t ca_context_cache                         = <ca_context_cache_t>dlsym(libcanberra, 'ca_context_cache')
    ca_context_cancel_t ca_context_cancel                       = <ca_context_cancel_t>dlsym(libcanberra, 'ca_context_cancel')
    ca_context_playing_t ca_context_playing                     = <ca_context_playing_t>dlsym(libcanberra, 'ca_context_playing')
    ca_strerror_t ca_strerror                                   = <ca_strerror_t>dlsym(libcanberra, 'ca_strerror')


class CanberraError(Exception):
    """Exception raised when a libcanberra API call returns an error"""

    def __init__(self, code: Union[int, Errors], msg: str = None):
        self.code = Errors(code)

        if msg is None:
            msg = ca_strerror(code).decode('utf-8')
        self.msg = msg

    def __str__(self):
        return f'{self.code!r}: {self.msg}'


cdef raise_if_error(int error):
    if error == CA_ERROR_OOM:
        raise MemoryError()

    if error != CA_SUCCESS:
        raise CanberraError(code=error)


cdef void ca_finish_callback(ca_context *ca, uint32_t id, int error_code, void *userdata) with gil:
    cdef PyGILState_STATE gilstate

    cdef PyObject **py_data = <PyObject **> userdata
    cdef object context = <object>py_data[0]
    cdef object callback = <object>py_data[1]
    cdef object callback_arg = <object>py_data[2]

    free(userdata)

    error = Errors(error_code)
    if error != Errors.SUCCESS:
        error = CanberraError(code=error)

    gilstate = PyGILState_Ensure()
    try:
        #
        # ca_finish_callbacks are not allowed to call libcanberra API functions,
        # as they may cause deadlocks (or fatal errors).
        #
        # We must be especially careful with DECREFing in the ca_finish_callback,
        # as well, for if the Context gets GC'd during this callback,
        # ca_context_destroy will be called on it, undoubtedly causing a fatal
        # error.
        #
        # So, instead of invoking user callbacks or DECREFing our context
        # directly from the ca_finish_callback, we queue these to be run in a
        # separate callback thread.
        #
        callback_queue.put_nowait((callback, context, id, error, callback_arg))
    finally:
        PyGILState_Release(gilstate)


cdef callback_processor():
    cdef object callback
    cdef object context
    cdef uint32_t id
    cdef object error
    cdef object callback_arg

    while True:
        callback, context, id, error, callback_arg = callback_queue.get()

        try:
            if callback is not None:
                if callback_arg is NOTSET:
                    args = (context, id, error)
                else:
                    args = (context, id, error, callback_arg)

                callback(*args)
        finally:
            Py_DECREF(context)
            Py_DECREF(callback)
            Py_DECREF(callback_arg)

            del context
            del callback
            del callback_arg
            del error


cdef callback_queue = Queue()
cdef callback_thread_canceled = Event()
cdef callback_thread = Thread(name='py-canberra callback thread', target=callback_processor, daemon=True)
callback_thread.start()


OnFinishedCallbackWithoutArg = Callable[['Context', int, Union[Errors, CanberraError]], Any]
OnFinishedCallbackWithArg = Callable[['Context', int, Union[Errors, CanberraError], Any], Any]
OnFinishedCallback = Union[OnFinishedCallbackWithArg, OnFinishedCallbackWithoutArg]


cdef populate_propslist(ca_proplist *proplist,
                        props: Dict[Union[str, Props], str],
                        other_props: Dict[Union[str, Props], str]):
    cdef bytes b_prop
    cdef char *c_prop
    cdef bytes b_value
    cdef char *c_value

    prop_values = Props.from_kwargs(props, **other_props)

    for prop, value in prop_values.items():
        #
        # libcanberra property names need to be in 7bit ASCII, string
        # property values UTF8.
        #
        b_prop = str(prop).encode('ascii')
        c_prop = b_prop

        b_value = str(value).encode('utf-8')
        c_value = b_value

        error = ca_proplist_sets(proplist, c_prop, c_value)
        raise_if_error(error)


cdef class Context:
    """A libcanberra ``ca_context``"""

    cdef ca_context *_ca_ctx

    def __cinit__(self):
        self._ca_ctx = NULL

        error = ca_context_create(&self._ca_ctx)
        if self._ca_ctx is NULL:
            raise MemoryError()

        raise_if_error(error)

    def __dealloc__(self):
        if self._ca_ctx is not NULL:
            ca_context_destroy(self._ca_ctx)

    def __init__(self, props: Dict[Union[str, Props], str] = None, **other_props: str):
        """Initialize the libcanberra ca_context, optionally with default props all sounds will share

        :param props:
            A mapping of Props to values. These properties and values will be set
            as defaults for all sounds played from this context.

        :param other_props:
            Props may also be passed as kwargs, where ``{Props.EVENT_ID: 'bell'}`` may
            be passed as ``event_id='bell'``. These properties and values will be set
            as defaults for all sounds played from this context.

        """
        if props or other_props:
            self.change_props(props, **other_props)

    def set_driver(self, driver: Union[str, bytes]) -> None:
        """Specify the backend driver used

        This method may not be called again after a successful call to :meth:`open`,
        which occurs implicitly when calling :meth:`.play`.

        This method might succeed even when the specified driver backend is not
        available. Use :meth:`open` to find out whether the backend is available.

        :param driver:
            The backend driver to use (e.g. ``"alsa"``, ``"pulse"``, ``"null"``, ...)

        """
        driver_bytes = driver if isinstance(driver, bytes) else driver.encode('utf-8')
        cdef char *c_driver = driver_bytes

        cdef int error = ca_context_set_driver(self._ca_ctx, c_driver)
        raise_if_error(error)

    def change_device(self, device: Union[str, bytes]) -> None:
        """Specify the backend device to use

        This method may not be called again after a successful call to :meth:`open`,
        which occurs implicitly when calling :meth:`.play`.

        This method might succeed even when the specified driver backend is not
        available. Use :meth:`open` to find out whether the backend is available.

        Depending on the backend used, this might or might not cause all
        currently playing event sounds to be moved to the new device.

        :param device:
            The backend device to use, in a format that is specific to the backend

        """
        device_bytes = device if isinstance(device, bytes) else device.encode('utf-8')
        cdef char *c_device = device_bytes

        cdef int error = ca_context_change_device(self._ca_ctx, c_device)
        raise_if_error(error)

    def open(self) -> None:
        """Connect the context to the sound system.

        This is implicitly called in :meth:`.play` or :meth:`cache` if not called
        explicitly. It is recommended to initialize application properties
        with :meth:`change_props` (or when creating :class:`Context`)
        before calling this function.

        """
        cdef int error = ca_context_open(self._ca_ctx)
        raise_if_error(error)

    def change_props(self, props: Dict[Union[str, Props], str] = None, **other_props: str) -> None:
        """Write one or more string properties to the Context

        Properties set like this will be attached to both the client object of
        the sound server and to all event sounds played or cached. It is
        recommended to call this method at least once before calling :meth:`open`
        (which occurs implicitly when calling :meth:`.play`), so that the
        initial application properties are set properly before the initial
        connection to the sound system.

        This method can be called both before and after the :meth:`open` call.
        Properties that have already been set before will be overwritten.

        :param props:
            A mapping of :class:`Props` to values.

        :param other_props:
            :class:`Props` may also be passed as kwargs, where
            ``{Props.EVENT_ID: 'bell'}`` is passed as ``event_id='bell'``.

        """
        cdef int error
        cdef ca_proplist *proplist

        error = ca_proplist_create(&proplist)
        raise_if_error(error)

        try:
            populate_propslist(proplist, props, other_props)

            error = ca_context_change_props_full(self._ca_ctx, proplist)
            raise_if_error(error)

        finally:
            ca_proplist_destroy(proplist)

    def cache(self, props: Dict[Union[str, Props], str] = None, **other_props: str) -> None:
        """Upload the specified sample into the audio server and attach the specified properties to it

        This method will only return after the sample upload was finished.

        The sound to cache is found with the same algorithm that is used to
        find the sounds for :meth:`.play`.

        If the backend doesn't support caching sound samples, this method
        will raise a :exc:`CanberraError` with a ``code`` of :attr:`.NOTSUPPORTED`.

        :param props:
            A mapping of :class:`Props` to values.

        :param other_props:
            :class:`Props` may also be passed as kwargs, where
            ``{Props.EVENT_ID: 'bell'}`` is passed as ``event_id='bell'``.

        """
        cdef int error
        cdef ca_proplist *proplist

        error = ca_proplist_create(&proplist)
        raise_if_error(error)

        try:
            populate_propslist(proplist, props, other_props)

            error = ca_context_cache_full(self._ca_ctx, proplist)
            raise_if_error(error)

        finally:
            ca_proplist_destroy(proplist)

    def play(
        self,
        props: Dict[Union[str, Props], str] = None,
        uint32_t id = 0,
        on_finished: OnFinishedCallback = None,
        user_data = NOTSET,
        **other_props: str,
    ) -> None:
        """Play one event sound

        ``id`` can be any numeric value which later can be used to cancel an
        event sound that is currently being played. You may use the same ``id``
        twice or more times if you want to cancel multiple event sounds with
        a single :meth:`cancel` call. It is recommended to pass ``0`` (the default)
        for the ``id`` if the event sound shall never be canceled.

        If the requested sound is not cached in the server yet, this call might
        result in the sample being uploaded temporarily or permanently
        (this may be controlled with :attr:`.CANBERRA_CACHE_CONTROL`).

        This method will start playback in the background. It will not wait to
        return until playback completed. Depending on the backend used, a sound
        that is started shortly before your application terminates might or
        might not continue to play after your application terminated. If you
        want to make sure that all sounds finish to play, you need to pass an
        :paramref:`on_finished <Context.play.on_finished>` callback and wait
        for it to be called before you terminate your application.

        The sample to play is identified by the :attr:`.EVENT_ID` property.
        If it is already cached in the server the cached version is played.
        The properties passed in this call are merged with the properties
        supplied when the sample was cached (if applicable) and the context
        properties as set with :meth:`change_props`.

        If :attr:`.EVENT_ID` is not defined, the sound file passed in the
        :attr:`.MEDIA_FILENAME` is played.

        On Linux/Unix, the right sound to play is determined according to
        :attr:`.EVENT_ID`, :attr:`.APPLICATION_LANGUAGE`/:attr:`.MEDIA_LANGUAGE`,
        the system locale, :attr:`.CANBERRA_XDG_THEME_NAME` and
        :attr:`.CANBERRA_XDG_THEME_OUTPUT_PROFILE`, following the XDG Sound
        Theming Specification. On non-Unix systems, the native event sound
        that matches the XDG sound name in :attr:`.EVENT_ID` is played.

        :param props:
            A mapping of :class:`Props` to values describing additional
            properties for this sound event.

        :param id:
            An integer id this sound can later be identified with when calling
            :meth:`.cancel`

        :param on_finished:
            An optional callback to be called when the sound finishes playing.
            Depending on whether :paramref:`.user_data` is passed, the prototype
            for this callback will be:

            .. code-block:: python

                # without user_data
                def callback(ctx: Context, id: int, error: Union[CanberraError, Errors])

                # with user_data
                def callback(ctx: Context, id: int, error: Union[CanberraError, Errors], user_data: Any)

        :param user_data:
            An optional argument to be passed to the
            :paramref:`on_finished <Context.play.on_finished>` callback

        :param other_props:
            :class:`Props` may also be passed as kwargs, where ``{Props.EVENT_ID: 'bell'}``
            is passed as ``event_id='bell'``, which describe additional
            properties for this sound event.

        """
        cdef int error
        cdef ca_proplist *proplist
        cdef PyObject **ca_userdata

        error = ca_proplist_create(&proplist)
        raise_if_error(error)

        try:
            populate_propslist(proplist, props, other_props)

            ca_userdata = <PyObject **>malloc(sizeof(PyObject *) * 3)
            ca_userdata[0] = <PyObject *>self
            ca_userdata[1] = <PyObject *>on_finished
            ca_userdata[2] = <PyObject *>user_data

            Py_INCREF(self)
            Py_INCREF(on_finished)
            Py_INCREF(user_data)

            error = ca_context_play_full(self._ca_ctx, id, proplist, ca_finish_callback, <void *>ca_userdata)
            if error != CA_SUCCESS:
                #
                # If the call is not successful, our ca_finish_callback will not
                # be called, so we must free our memory and decrement our references
                # now.
                #
                free(ca_userdata)

                Py_DECREF(self)
                Py_DECREF(on_finished)
                Py_DECREF(user_data)

            raise_if_error(error)

        finally:
            ca_proplist_destroy(proplist)

    def cancel(self, uint32_t id = 0) -> None:
        """Cancel one or more event sounds that have been started via :meth:`.play`

        If callback function was passed to :meth:`.play` when starting the sound,
        calling :meth:`.cancel` might cause this callback function to be called
        with :attr:`.CANCELED` as the error code (wrapped in :exc:`CanberraError`).

        :param id:
            The ID that identifies the sound(s) to cancel.

        """
        cdef int error

        error = ca_context_cancel(self._ca_ctx, id)
        raise_if_error(error)

    def playing(self, uint32_t id = 0) -> bool:
        """Check if at least one sound with the specified id is still playing

        :param id:
            The ID that identifies the sound(s) to check

        :return:
            ``True`` if at least one sound with the specified ID is still playing,
            and ``False`` if no sounds with the specified ID are still playing

        """
        cdef int error
        cdef int playing

        error = ca_context_playing(self._ca_ctx, id, &playing)
        raise_if_error(error)

        return bool(playing)



from typing import Any, Callable, Dict, Union

from canberra._canberra import CanberraError
from canberra.constants import Errors, Props, NOTSET

OnFinishedCallbackWithoutArg = Callable[['Context', int, Union[Errors, CanberraError]], Any]
OnFinishedCallbackWithArg = Callable[['Context', int, Union[Errors, CanberraError], Any], Any]
OnFinishedCallback = Union[OnFinishedCallbackWithArg, OnFinishedCallbackWithoutArg]


class Context:
    def __init__(self, props: Dict[Union[str, Props], str] = None, **other_props: str): ...
    def set_driver(self, driver: Union[str, bytes]) -> None: ...
    def change_device(self, device: Union[str, bytes]) -> None: ...
    def open(self) -> None: ...
    def change_props(self, props: Dict[Union[str, Props], str] = None, **other_props: str) -> None: ...
    def cache(self, props: Dict[Union[str, Props], str] = None, **other_props: str) -> None: ...
    def play(
        self,
        props: Dict[Union[str, Props], str] = None,
        id: int = 0,
        on_finished: OnFinishedCallback = None,
        user_data=NOTSET,
        **other_props: str,
    ) -> None: ...
    def cancel(self, id: int = 0) -> None: ...
    def playing(self, id: int = 0) -> bool: ...



from enum import Enum, IntEnum
from typing import Dict, Union

NOTSET = object()


class StrEnum(str, Enum):
    def __str__(self):
        return self.value


class Props(StrEnum):
    """Properties that describe sounds being played

    .. note::

        Any methods that accept :class:`Props` values will also accept them as
        kwargs, using the lowercased attribute name

        .. code-block:: python

            ctx = Context({Props.APPLICATION_NAME: 'My Application'})
            # or
            ctx = Context(application_name='My Application')

    """

    MEDIA_NAME = 'media.name'
    """
    A name describing the media being played. Localized if possible and applicable.
    """

    MEDIA_TITLE = 'media.title'
    """
    A (song) title describing the media being played. Localized if possible and applicable.
    """

    MEDIA_ARTIST = 'media.artist'
    """
    The artist of this media. Localized if possible and applicable.
    """

    MEDIA_LANGUAGE = 'media.language'
    """
    The language this media is in, in some standard POSIX locale string,
    such as ``"de_DE"``.
    """

    MEDIA_FILENAME = 'media.filename'
    """
    The file name this media was or can be loaded from.
    """

    MEDIA_ICON = 'media.icon'
    """
    An icon for this media in binary PNG format.
    """

    MEDIA_ICON_NAME = 'media.icon_name'
    """
    An icon name as defined in the XDG icon naming specification.
    """

    MEDIA_ROLE = 'media.role'
    """
    The "role" this media is played in. For event sounds the string
    ``"event"``. For other cases, strings like ``"music"``, ``"video"``,
    ``"game"``, ...
    """

    EVENT_ID = 'event.id'
    """
    A textual id for an event sound, as mandated by the XDG sound naming specification.
    """

    EVENT_DESCRIPTION = 'event.description'
    """
    A descriptive string for the sound event. Localized if possible and applicable.
    """

    EVENT_MOUSE_X = 'event.mouse.x'
    """
    If this sound event was triggered by a mouse input event, the X
    position of the mouse cursor on the screen, formatted as string.
    """

    EVENT_MOUSE_Y = 'event.mouse.y'
    """
    If this sound event was triggered by a mouse input event, the Y
    position of the mouse cursor on the screen, formatted as string.
    """

    EVENT_MOUSE_HPOS = 'event.mouse.hpos'
    """
    If this sound event was triggered by a mouse input event, the X
    position of the mouse cursor as fractional value between ``0`` and ``1``,
    formatted as string, ``0`` reflecting the left side of the screen, ``1``
    the right side.
    """

    EVENT_MOUSE_VPOS = 'event.mouse.vpos'
    """
    If this sound event was triggered by a mouse input event, the Y
    position of the mouse cursor as fractional value between ``0`` and ``1``,
    formatted as string, ``0`` reflecting the top end of the screen, ``1``
    the bottom end.
    """

    EVENT_MOUSE_BUTTON = 'event.mouse.button'
    """
    If this sound event was triggered by a mouse input event, the
    number of the mouse button that triggered it, formatted as string. ``1``
    for left mouse button, ``3`` for right, ``2`` for middle.
    """

    WINDOW_NAME = 'window.name'
    """
    If this sound event was triggered by a window on the screen, the
    name of this window as human readable string.
    """

    WINDOW_ID = 'window.id'
    """
    If this sound event was triggered by a window on the screen, some
    identification string for this window, so that the sound system can
    recognize specific windows.
    """

    WINDOW_ICON = 'window.icon'
    """
    If this sound event was triggered by a window on the screen, binary
    icon data in PNG format for this window.
    """

    WINDOW_ICON_NAME = 'window.icon_name'
    """
    If this sound event was triggered by a window on the screen, an
    icon name for this window, as defined in the XDG icon naming
    specification.
    """

    WINDOW_X = 'window.x'
    """
    If this sound event was triggered by a window on the screen, the X
    position of the window measured from the top left corner of the
    screen to the top left corner of the window.

    Since: 0.17
    """

    WINDOW_Y = 'window.y'
    """
    If this sound event was triggered by a window on the screen, the y
    position of the window measured from the top left corner of the
    screen to the top left corner of the window.

    Since: 0.17
    """

    WINDOW_WIDTH = 'window.width'
    """
    If this sound event was triggered by a window on the screen, the
    pixel width of the window.

    Since: 0.17
    """

    WINDOW_HEIGHT = 'window.height'
    """
    If this sound event was triggered by a window on the screen, the
    pixel height of the window.

    Since: 0.17
    """

    WINDOW_HPOS = 'window.hpos'
    """
    If this sound event was triggered by a window on the screen, the X
    position of the center of the window as fractional value between 0
    and 1, formatted as string, 0 reflecting the left side of the
    screen, 1 the right side.

    Since: 0.17
    """

    WINDOW_VPOS = 'window.vpos'
    """
    If this sound event was triggered by a window on the screen, the Y
    position of the center of the window as fractional value between 0
    and 1, formatted as string, 0 reflecting the top side of the
    screen, 1 the bottom side.

    Since: 0.17
    """

    WINDOW_DESKTOP = 'window.desktop'
    """
    If this sound event was triggered by a window on the screen and the
    windowing system supports multiple desktops, a comma separated list
    of indexes of the desktops this window is visible on. If this
    property is an empty string, it is visible on all desktops
    (i.e. 'sticky'). The first desktop is 0. (e.g. ``"0,2,3"``)

    Since: 0.18
    """

    WINDOW_X11_DISPLAY = 'window.x11.display'
    """
    If this sound event was triggered by a window on the screen and the
    windowing system is X11, the X display name of the window (e.g. ``":0"``).
    """

    WINDOW_X11_SCREEN = 'window.x11.screen'
    """
    If this sound event was triggered by a window on the screen and the
    windowing system is X11, the X screen id of the window formatted as
    string (e.g. ``"0"``).
    """

    WINDOW_X11_MONITOR = 'window.x11.monitor'
    """
    If this sound event was triggered by a window on the screen and the
    windowing system is X11, the X monitor id of the window formatted as
    string (e.g. ``"0"``).
    """

    WINDOW_X11_XID = 'window.x11.xid'
    """
    If this sound event was triggered by a window on the screen and the
    windowing system is X11, the XID of the window formatted as string.
    """

    APPLICATION_NAME = 'application.name'
    """
    The name of the application this sound event was triggered by as
    human readable string (e.g. ``"GNU Emacs"``). Localized if possible and
    applicable.
    """

    APPLICATION_ID = 'application.id'
    """
    An identifier for the program this sound event was triggered
    by. (e.g. "org.gnu.emacs").
    """

    APPLICATION_VERSION = 'application.version'
    """
    A version number for the program this sound event was triggered
    by. (e.g. ``"22.2"``)
    """

    APPLICATION_ICON = 'application.icon'
    """
    Binary icon data in PNG format for the application this sound event
    is triggered by.
    """

    APPLICATION_ICON_NAME = 'application.icon_name'
    """
    An icon name for the application this sound event is triggered by,
    as defined in the XDG icon naming specification.
    """

    APPLICATION_LANGUAGE = 'application.language'
    """
    The locale string the application that is triggering this sound
    event is running in. A POSIX locale string such as ``de_DE@euro``.
    """

    APPLICATION_PROCESS_ID = 'application.process.id'
    """
    The unix PID of the process that is triggering this sound event, formatted as string.
    """

    APPLICATION_PROCESS_BINARY = 'application.process.binary'
    """
    The path to the process binary of the process that is triggering this sound event.
    """

    APPLICATION_PROCESS_USER = 'application.process.user'
    """
    The user that owns the process that is triggering this sound event.
    """

    APPLICATION_PROCESS_HOST = 'application.process.host'
    """
    The host name of the host the process that is triggering this sound event runs on.
    """

    CANBERRA_CACHE_CONTROL = 'canberra.cache-control'
    """
    A special property that can be used to control the automatic sound
    caching of sounds in the sound server. One of ``"permanent"``,
    ``"volatile"``, ``"never"``. ``"permanent"`` will cause this sample to be
    cached in the server permanently. This is useful for very frequently used
    sound events such as those used for input feedback. ``"volatile"`` may be
    used for caching sounds in the sound server temporarily. They will expire
    after some time or on cache pressure. Finally, ``"never"`` may be used for
    sounds that should never be cached, because they are only generated very
    seldomly or even only once at most (such as desktop login sounds).

    If this property is not explicitly passed to :meth:`Context.play()` it
    will default to ``"never"``. If it is not explicitly passed to
    :meth:`Context.cache()` it will default to "permanent".

    If the list of properties is handed on to the
    sound server this property is stripped from it.
    """

    CANBERRA_VOLUME = 'canberra.volume'
    """
    A special property that can be used to control the volume this
    sound event is played in if the backend supports it. A floating
    point value for the decibel multiplier for the sound. 0 dB relates
    to zero gain, and is the default volume these sounds are played in.

    NOTE: the value of this must be in decibels, not in amplitudes,
          e.g., from 0.0 to 1.0.

    If the list of properties is handed on to the sound server this
    property is stripped from it.
    """

    CANBERRA_XDG_THEME_NAME = 'canberra.xdg-theme.name'
    """
    A special property that can be used to control the XDG sound theme that
    is used for this sample.

    If the list of properties is handed on to the sound server this
    property is stripped from it.
    """

    CANBERRA_XDG_THEME_OUTPUT_PROFILE = 'canberra.xdg-theme.output-profile'
    """
    A special property that can be used to control the XDG sound theme
    output profile that is used for this sample.

    If the list of properties is handed on to the sound server this
    property is stripped from it.
    """

    CANBERRA_ENABLE = 'canberra.enable'
    """
    A special property that can be used to control whether any sounds
    are played at all. If this property is ``"1"`` or unset sounds are
    played as normal. However, if it is ``"0"`` all calls to
    :meth:`Context.play()` will fail with :attr:`.DISABLED`.

    If the list of properties is handed on to the sound server this
    property is stripped from it.
    """

    CANBERRA_FORCE_CHANNEL = 'canberra.force_channel'
    """
    A special property that can be used to control on which channel a
    sound is played. The value should be one of ``mono``, ``front-left``,
    ``front-right``, ``front-center``, ``rear-left``, ``rear-right``,
    ``rear-center``, ``lfe``, ``front-left-of-center``, ``front-right-of-center``,
    ``side-left``, ``side-right``, ``top-center``, ``top-front-left``,
    ``top-front-right``, ``top-front-center``, ``top-rear-left``, ``top-rear-right``,
    ``top-rear-center``.

    This property is only honoured by some backends, other backends may choose
    to ignore it completely.

    If the list of properties is handed on to the sound server this
    property is stripped from it.

    Since: 0.13
    """

    ###
    # A mapping of lowercase names to the Props enum values they represent.
    # This mapping is used to power canberra.Context().play(media_id='...'), etc
    #
    kwarg_names: Dict[str, 'Props']

    @classmethod
    def from_kwargs(cls,
                    props: Dict[Union[str, 'Props'], str] = None,
                    **kwargs: str,
                    ) -> Dict['Props', str]:
        """Translate kwarg names to Props for a Props->value mapping
        """
        inputs = {}
        if props:
            inputs.update(props)
        inputs.update(kwargs)

        prop_values: Dict['Props', str] = {}
        for key, value in inputs.items():
            if isinstance(key, str):
                if key in cls.kwarg_names:
                    key = cls.kwarg_names[key]
                else:
                    key = cls(key)

            if not isinstance(key, cls):
                raise TypeError(f'Expected a valid {type(key)}. Found {type(key)} {key!r}')

            prop_values[key] = value

        return prop_values


Props.kwarg_names = {
    k.lower(): v
    for k, v in Props.__members__.items()
}


class Errors(IntEnum):
    """libcanberra error codes

    Negative numbers represent errors; ``0`` represents success
    """

    SUCCESS = 0
    """Success"""

    NOTSUPPORTED = -1
    """Operation not supported"""

    INVALID = -2
    """Invalid argument"""

    STATE = -3
    """Invalid state"""

    OOM = -4
    """Out of memory"""

    NODRIVER = -5
    """No such driver"""

    SYSTEM = -6
    """System error"""

    CORRUPT = -7
    """File or data corrupt"""

    TOOBIG = -8
    """File or data too large"""

    NOTFOUND = -9
    """File or data not found"""

    DESTROYED = -10
    """Destroyed"""

    CANCELED = -11
    """Canceled"""

    NOTAVAILABLE = -12
    """Not available"""

    ACCESS = -13
    """Access forbidden"""

    IO = -14
    """IO error"""

    INTERNAL = -15
    """Internal error"""

    DISABLED = -16
    """Sound disabled"""

    FORKED = -17
    """Process forked"""

    DISCONNECTED = -18
    """Disconnected"""



import re
from setuptools import Extension, setup

from Cython.Build import cythonize


def get_readme():
    with open('README.rst') as fp:
        return fp.read()


def get_version():
    with open('canberra/__init__.py') as fp:
        match = re.search(r"__version__ = '([^']+)'", fp.read())
        return match.group(1)


setup(
    long_description=get_readme(),
    long_description_content_type='text/x-rst',
    name='py-canberra',
    version=get_version(),
    description='Wrappers for the libcanberra sound-playing interface',
    url='https://github.com/theY4Kman/py-canberra',
    python_requires='==3.*,>=3.6.0',
    project_urls={
        'Repository': 'https://github.com/theY4Kman/py-canberra',
        'Documentation': 'https://py-canberra.readthedocs.io/',
    },
    author='Zach "theY4Kman" Kanzler',
    author_email='they4kman@gmail.com',
    license='MIT',
    packages=['canberra'],
    package_data={'canberra': ['_canberra.pyi']},
    extras_require={"dev": ["cython==0.*,>=0.29.19"]},
    ext_modules=cythonize([
        Extension(
            name='canberra._canberra',
            sources=['canberra/_canberra.pyx'],
        ),
    ]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Multimedia :: Sound/Audio',
    ],
)



__version__ = '0.0.4'

from .constants import Props, Errors
from ._canberra import Context
from .convenience import play, play_file


__all__ = [
    'Context',
    'Props',
    'Errors',
    'play',
    'play_file',
]
